#! /usr/bin/env python
"""Dependencies for RM utilities """
__all__ = ['mpfit',
           'normalize',
           'util_FITS',
           'util_misc',
           'util_plotFITS',
           'util_plotTk',
           'util_rec',
           'util_RM']
